#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Anchorage module.

Implements:

* class WorldAnchorage2D -- geometrical anchorage for 2D world.
* class WorldAnchorage3D -- geometrical anchorage for 3D world.
"""

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

try:
    from .point import *
    from .vector import *
except:
    from point import *
    from vector import *
# end try


class WorldAnchorage2D:
    """Geometrical anchorage for 2D world."""

    # constant defs
    C = CENTER = "c"
    E = EAST = "e"
    N = NORTH = "n"
    S = SOUTH = "s"
    W = WEST = "w"

    LEGAL_ANCHORS = (C, E, N, N+E, N+W, S, S+E, S+W, W)

    DEFAULT_ANCHOR = CENTER
    DEFAULT_ORIGIN = None
    DEFAULT_UNIT_VECTOR = (1, 1)

    def __init__ (self, **options):
        """Class constructor.

        Parameter:
            options -- keyword arguments.

        Supported keyword arguments:
            anchor -- anchorage query (default: CENTER).

            origin -- anchorage point of origin (default: None).

            unit_vector -- local unit vector (default: (1, 1)).

        Notice: options are saved into self.options member attribute.
        """
        self.options = options
        self.anchor = options.get("anchor")
        self.origin = options.get("origin")
        self.unit_vector = options.get("unit_vector")
    # end def

    @property
    def anchor (self):
        """Anchorage query - read-write attribute.

        This must be one of self.LEGAL_ANCHORS e.g. CENTER, EAST,
        NORTH, etc.
        """
        return self.__anchor
    # end def

    @anchor.setter
    def anchor (self, value):
        value = str(value or self.DEFAULT_ANCHOR).lower()
        if value in self.LEGAL_ANCHORS:
            self.__anchor = value
        else:
            raise ValueError(
                "anchor should be one of {}."
                .format(self.LEGAL_ANCHORS)
            )
        # end if
    # end def

    @property
    def origin (self):
        """Anchorage point of origin - read-write attribute."""
        return self.__origin
    # end def

    @origin.setter
    def origin (self, value):
        self.__origin = self.point_type(value or self.DEFAULT_ORIGIN)
    # end def

    @property
    def point_type (self):
        """Preferred point type - read-only attribute.

        Current implementation returns WorldPoint2D class type.

        This can be redefined in subclasses to fit more specific needs.
        """
        return WorldPoint2D
    # end def

    @property
    def unit_vector (self):
        """Local unit vector - read-write attribute."""
        return self.__unit_vector
    # end def

    @unit_vector.setter
    def unit_vector (self, value):
        self.__unit_vector = self.vector_type(
            value or self.DEFAULT_UNIT_VECTOR
        )
    # end def

    @property
    def vector_type (self):
        """Preferred vector type - read-only attribute.

        Current implementation returns WorldVector2D class type.

        This can be redefined in subclasses to fit more specific needs.
        """
        return WorldVector2D
    # end def

# end class


class WorldAnchorage3D (WorldAnchorage2D):
    """Geometrical anchorage for 3D world."""

    # constant defs
    F = FRONT = "f"
    M = MIDDLE = "m"
    R = REAR = "r"

    LEGAL_ANCHORS = tuple(
        a + b
        for a in (F, M, R)
        for b in WorldAnchorage2D.LEGAL_ANCHORS
    )

    DEFAULT_ANCHOR = MIDDLE + WorldAnchorage2D.CENTER
    DEFAULT_ORIGIN = None
    DEFAULT_UNIT_VECTOR = (1, 1, 1)

    @property
    def point_type (self):
        """Preferred point type - read-only attribute.

        Current implementation returns WorldPoint3D class type.

        This can be redefined in subclasses to fit more specific needs.
        """
        return WorldPoint3D
    # end def

    @property
    def vector_type (self):
        """Preferred vector type - read-only attribute.

        Current implementation returns WorldVector3D class type.

        This can be redefined in subclasses to fit more specific needs.
        """
        return WorldVector3D
    # end def

# end class
