#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Playground area module.

Implements:

* class WorldArea2D -- playground area for 2D world.
* class WorldArea3D -- playground area for 3D world.
"""

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

try:
    from .box_base import *
    from .bounding_box import *
except:
    from box_base import *
    from bounding_box import *
# end try


class WorldArea2D (WorldBoxBase2D):
    """Playground area for 2D world."""

    def __init__ (self, width=0, height=0, **options):
        """Class constructor.

        Parameters:
            width -- area's width (default: 0).

            height -- area's height (default: 0).

            options -- keyword arguments.

        Supported keyword arguments:
            bounds or limits -- a couple of tuple coordinates or point
            objects of diagonal vertices of area's bounding box e.g.
            ((0, 0), (800, 600)). Defaults to self.default_bounds if
            None or omitted.

        Notice: options are saved into self.options member attribute.
        """
        super().__init__(width, height, **options)
        self.bounds = options.get("bounds") or options.get("limits")
    # end def

    @property
    def bounds (self):
        """Area's bounding box - read-write attribute."""
        return self.__bounds
    # end def

    @bounds.setter
    def bounds (self, value):
        self.__bounds = self.bounds_type(
            *(value or self.default_bounds)
        )
    # end def

    @property
    def bounds_type (self):
        """Preferred bounding box type - read-only attribute.

        Current implementation returns WorldBoundingBox2D class type.

        This can be redefined in subclasses to fit more specific needs.
        """
        return WorldBoundingBox2D
    # end def

    @property
    def default_bounds (self):
        """Default bounds for bounding box - read-only attribute."""
        return ((0,) * len(self.size), self.size)
    # end def

# end class


class WorldArea3D (WorldBoxBase3D, WorldArea2D):
    """Playground area for 3D world."""

    def __init__ (self, width=0, height=0, depth=0, **options):
        """Class constructor.

        Parameters:
            width -- area's width (default: 0).

            height -- area's height (default: 0).

            depth -- area's depth (default: 0).

            options -- keyword arguments.

        Supported keyword arguments:
            bounds or limits -- a couple of tuple coordinates or point
            objects of diagonal vertices of area's bounding box e.g.
            ((0, 0, 0), (800, 600, 400)). Defaults to
            self.default_bounds if None or omitted.

        Notice: options are saved into self.options member attribute.
        """
        WorldBoxBase3D.__init__(self, width, height, depth, **options)
        self.bounds = options.get("bounds") or options.get("limits")
    # end def

    @property
    def bounds_type (self):
        """Preferred bounding box type - read-only attribute.

        Current implementation returns WorldBoundingBox3D class type.

        This can be redefined in subclasses to fit more specific needs.
        """
        return WorldBoundingBox3D
    # end def

# end class
