#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Bounding box module.

Implements:

* class WorldBoundingBox2D -- bounding box for 2D world.
* class WorldBoundingBox3D -- bounding box for 3D world.
"""

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import itertools

try:
    from .point import *
except:
    from point import *
# end try


class WorldBoundingBox2D:
    """Bounding box for 2D world."""

    def __getitem__ (self, key):
        """Make object iterable."""
        return self.to_tuple()[key]
    # end def

    def __init__ (self, *points, **options):
        """Class constructor.

        Parameters:
            points -- variable list of tuple coordinates or point
            objects for bounds calculations.

            options -- keyword arguments.

        Supported keyword arguments:
            -- none at this stage --

        Notice: options are saved into self.options member attribute.
        """
        self.reset_bounds(*points)
        self.options = options
    # end def

    def __repr__ (self):
        """Object string representation."""
        return str(self.to_tuple())
    # end def

    def bounding_coords (self, *points):
        """Bounding coordinates.

        Parameter:
            points -- variable list of tuple coordinates or point
            objects.

        Returns a 2-tuple of tuple coordinates - minimum and maximum
        coordinates along with given points.
        """
        _coords = tuple(zip(*filter(None, points)))
        return (tuple(map(min, _coords)), tuple(map(max, _coords)))
    # end def

    def bounding_corners (self, *points):
        """Bounding box corner points.

        Parameter:
            points -- variable list of tuple coordinates or point
            objects.

        returns a tuple list containing all point objects of bounding
        box' corners (2D --> 4 corners, 3D --> 8 corners).
        """
        _cmin, _cmax = self.bounding_coords(*points)
        return tuple(
            self.point_type(*coords)
            for coords in itertools.product(*zip(_cmin, _cmax))
        )
    # end def

    def bounding_points (self, *points):
        """Bounding box' min/max points.

        Parameter:
            points -- variable list of tuple coordinates or point
            objects.

        returns a 2-tuple list of bounding box' minimum coords point
        object and maximum coords point object.
        """
        points = self.bounding_coords(
            *(points or (self.point1, self.point2))
        )
        return tuple(self.point_type(*coords) for coords in points)
    # end def

    def center (self, query=None):
        """Build tuple of center dimensions along with query.

        Parameter:
            query -- string of dims (default: None).

        If None or omitted, `query` will be replaced by "xy" string of
        chars.

        Any combination of dims is allowed e.g. "x", "xy", "yx", "y".

        Will always return a tuple object, even for only one dimension
        query e.g. self.center("x") --> (self.center_x, ).
        """
        query = str(query or "xy").lower()
        _result = []
        if "x" in query:
            _result.append(self.center_x)
        # end if
        if "y" in query:
            _result.append(self.center_y)
        # end if
        return tuple(_result)
    # end def

    @property
    def center_x (self):
        """Center of X dimension - read-only attribute.

        Current implementation merely returns
        (self.point1.x + self.point2.x)//2.
        """
        return (self.point1.x + self.point2.x)//2
    # end def

    @property
    def center_xy (self):
        """Tuple center of (X, Y) dimensions - read-only attribute.

        Current implementation merely returns a tuple of
        (self.center_x, self.center_y).
        """
        return (self.center_x, self.center_y)
    # end def

    @property
    def center_y (self):
        """Center of Y dimension - read-only attribute.

        Current implementation merely returns
        (self.point1.y + self.point2.y)//2.
        """
        return (self.point1.y + self.point2.y)//2
    # end def

    def collided (self, *points):
        """Detect if any point is in current bounding box' bounds.

        Parameter:
            points -- variable list of tuple coordinates or point
            objects.

        Returns True if at least one of listed tuple coordinates (or
        point object) is into current bounding box' bounds.
        """
        return any(map(self.in_bounds, points))
    # end def

    @property
    def corners (self):
        """Default bounding box corner points - read-only attribute.

        See self.bounding_corners() docstring for more detail.
        """
        return self.bounding_corners(self.point1, self.point2)
    # end def

    def entered (self, *points):
        """Evaluate bounding box' entered sides.

        Parameter:
            points -- variable list of tuple coordinates or point
            objects.

        Returns a tuple list of any current bounding box side name
        ('left', 'right', 'top', 'bottom') that has been entered by
        given group of points.
        """
        _bounds = []
        _cmin, _cmax = self.bounding_coords(*points)
        _p_min_in = self.truth_point(_cmin)
        _p_max_in = self.truth_point(_cmax)
        _any_in = _p_min_in + _p_max_in
        if _any_in.y:
            if _p_max_in.x:
                _bounds.append("left")
            # end if
            if _p_min_in.x:
                _bounds.append("right")
            # end if
        # end if
        if _any_in.x:
            if _p_min_in.y:
                _bounds.append("top")
            # end if
            if _p_max_in.y:
                _bounds.append("bottom")
            # end if
        # end if
        return tuple(_bounds)
    # end def

    def exited (self, *points):
        """Evaluate bounding box' exited sides.

        Parameter:
            points -- variable list of tuple coordinates or point
            objects.

        Returns a tuple list of any current bounding box side name
        ('left', 'right', 'top', 'bottom') that has been exited by
        given group of points.
        """
        _bounds = []
        _p_min, _p_max = self.bounding_points(*points)
        _b_min, _b_max = self.bounding_points(self.point1, self.point2)
        if _p_min.x <= _b_min.x:
            _bounds.append("left")
        # end if
        if _p_max.x >= _b_max.x:
            _bounds.append("right")
        # end if
        if _p_max.y >= _b_max.y:
            _bounds.append("top")
        # end if
        if _p_min.y <= _b_min.y:
            _bounds.append("bottom")
        # end if
        return tuple(_bounds)
    # end def

    @property
    def height (self):
        """Bounding box' height - read-only attribute."""
        return abs(self.point1.y - self.point2.y)
    # end def

    def in_bounds (self, point, *points):
        """Detect if point is in points' bounding box bounds.

        Parameters:
            point -- tuple coordinates or point object.

            points -- variable list of tuple coordinates or point
            objects.

        Returns True if given point is included into the bounding box
        of listed points (or into current bounding box, if points are
        omitted).
        """
        return all(self.truth_point(point, *points))
    # end def

    def overlapped (self, *points):
        """Detect if current bounding box is overlapped by points.

        Parameter:
            points -- variable list of tuple coordinates or point
            objects.

        Returns True if current bounding box is overlapped in any
        manner by the bounding box of listed points.
        """
        return bool(
            self.in_bounds(self.point1, *points)
            or self.in_bounds(self.point2, *points)
        )
    # end def

    @property
    def point1 (self):
        """Minimum coordinates point - read-write attribute."""
        return self.__point1
    # end def

    @point1.setter
    def point1 (self, value):
        self.__point1 = self.point_type(*value)
    # end def

    @property
    def point2 (self):
        """Maximum coordinates point - read-write attribute."""
        return self.__point2
    # end def

    @point2.setter
    def point2 (self, value):
        self.__point2 = self.point_type(*value)
    # end def

    @property
    def point_type (self):
        """Preferred point type - read-only attribute.

        Current implementation returns WorldPoint2D class type.

        This can be redefined in subclasses to fit more specific needs.
        """
        return WorldPoint2D
    # end def

    def reset_bounds (self, *points):
        """Reset current bounding box' bounds.

        Parameter:
            points -- variable list of tuple coordinates or point
            objects.
        """
        self.point1, self.point2 = self.bounding_coords(*points)
    # end def

    @property
    def size (self):
        """Bounding box size - read-only attribute.

        Returns (width, height) tuple.
        """
        return (self.width, self.height)
    # end def

    def to_dict (self):
        """Dictionary representation of object attributes.

        Returns dict(point1=self.point1, point2=self.point2).
        """
        return dict(point1=self.point1, point2=self.point2)
    # end def

    def to_flat_tuple (self):
        """Flat tuple representation.

        All coordinates are linearized into a single tuple.

        This converts a ((x1, y1), ..., (xn, yn)) tuple list of tuple
        coordinates to a single (x1, y1, ..., xn, yn) tuple of flat
        coordinates.
        """
        return tuple(dim for point in self.to_tuple() for dim in point)
    # end def

    def to_tuple (self):
        """Tuple representation of object attributes.

        Returns (self.point1, self.point2) tuple.
        """
        return (self.point1, self.point2)
    # end def

    def truth_point (self, point, *points):
        """Boolean point representation of bounds detection.

        Parameters:
            point -- tuple coordinates or point object.

            points -- variable list of tuple coordinates or point
            objects.

        Returns a point object where each dimension coordinate
        represents the testing truth value of `point` dimension
        comprised between min and max of listed `points` dimension.

        In other words, we get something like::

            Point(
                int(minX(points) <= point.x <= maxX(points)),
                int(minY(points) <= point.y <= maxY(points)),
                int(minZ(points) <= point.z <= maxZ(points))
            )

        If `points` are omitted, `point` is compared to current
        bounding box points (self.point1, self.point2).
        """
        _coords = tuple(point)
        _cmin, _cmax = self.bounding_coords(
            *(points or (self.point1, self.point2))
        )
        return self.point_type(
            *(
                int(dmin <= dim <= dmax)
                for dmin, dim, dmax in zip(_cmin, _coords, _cmax)
            )
        )
    # end def

    @property
    def width (self):
        """Bounding box' width - read-only attribute."""
        return abs(self.point1.x - self.point2.x)
    # end def

# end class


class WorldBoundingBox3D (WorldBoundingBox2D):
    """Bounding box for 3D world."""

    def center (self, query=None):
        """Build tuple of center dimensions along with query.

        Parameter:
            query -- string of dims (default: None).

        If None or omitted, `query` will be replaced by "xyz" string of
        chars.

        Any combination of dims is allowed e.g. "x", "xy", "xz", "yz",
        "xyz", "yxz", "zxy", "zyx" and so on.

        Will always return a tuple object, even for only one dimension
        query e.g. self.center("x") --> (self.center_x, ).
        """
        query = str(query or "xyz").lower()
        _result = []
        if "x" in query:
            _result.append(self.center_x)
        # end if
        if "y" in query:
            _result.append(self.center_y)
        # end if
        if "z" in query:
            _result.append(self.center_z)
        # end if
        return tuple(_result)
    # end def

    @property
    def center_xyz (self):
        """Tuple center of (X, Y, Z) dimensions - read-only attribute.

        Current implementation merely returns a tuple of
        (self.center_x, self.center_y, self.center_z).
        """
        return (self.center_x, self.center_y, self.center_z)
    # end def

    @property
    def center_z (self):
        """Center of Z dimension - read-only attribute.

        Current implementation merely returns
        (self.point1.z + self.point2.z)//2.
        """
        return (self.point1.z + self.point2.z)//2
    # end def

    @property
    def depth (self):
        """Bounding box' depth - read-only attribute."""
        return abs(self.point1.z - self.point2.z)
    # end def

    def entered (self, *points):
        """Evaluate bounding box' entered sides.

        Parameter:
            points -- variable list of tuple coordinates or point
            objects.

        Returns a tuple list of any current bounding box side name
        ('left', 'right', 'top', 'bottom', 'front', 'rear') that has
        been entered by given group of points.
        """
        _bounds = []
        _cmin, _cmax = self.bounding_coords(*points)
        _p_min_in = self.truth_point(_cmin)
        _p_max_in = self.truth_point(_cmax)
        _any_in = _p_min_in + _p_max_in
        if _any_in.y and _any_in.z:
            if _p_max_in.x:
                _bounds.append("left")
            # end if
            if _p_min_in.x:
                _bounds.append("right")
            # end if
        # end if
        if _any_in.x and _any_in.z:
            if _p_min_in.y:
                _bounds.append("top")
            # end if
            if _p_max_in.y:
                _bounds.append("bottom")
            # end if
        # end if
        if _any_in.x and _any_in.y:
            if _p_max_in.z:
                _bounds.append("front")
            # end if
            if _p_min_in.z:
                _bounds.append("rear")
            # end if
        # end if
        return tuple(_bounds)
    # end def

    def exited (self, *points):
        """Evaluate bounding box' exited sides.

        Parameter:
            points -- variable list of tuple coordinates or point
            objects.

        Returns a tuple list of any current bounding box side name
        ('left', 'right', 'top', 'bottom', 'front', 'rear') that has
        been exited by given group of points.
        """
        _bounds = []
        _p_min, _p_max = self.bounding_points(*points)
        _b_min, _b_max = self.bounding_points(self.point1, self.point2)
        if _p_min.x <= _b_min.x:
            _bounds.append("left")
        # end if
        if _p_max.x >= _b_max.x:
            _bounds.append("right")
        # end if
        if _p_max.y >= _b_max.y:
            _bounds.append("top")
        # end if
        if _p_min.y <= _b_min.y:
            _bounds.append("bottom")
        # end if
        if _p_min.z <= _b_min.z:
            _bounds.append("front")
        # end if
        if _p_max.z >= _b_max.z:
            _bounds.append("rear")
        # end if
        return tuple(_bounds)
    # end def

    @property
    def point_type (self):
        """Preferred point type - read-only attribute.

        Current implementation returns WorldPoint3D class type.

        This can be redefined in subclasses to fit more specific needs.
        """
        return WorldPoint3D
    # end def

    @property
    def size (self):
        """Bounding box size - read-only attribute.

        Returns (width, height, depth) tuple.
        """
        return (self.width, self.height, self.depth)
    # end def

# end class
