#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Box base module.

Implements:

* class WorldBoxBase2D -- base class for boxes in a 2D world.
* class WorldBoxBase3D -- base class for boxes in a 3D world.
"""

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""


class WorldBoxBase2D:
    """Base class for boxes in a 2D world."""

    def __getitem__ (self, key):
        """Make object iterable."""
        return self.to_tuple()[key]
    # end def

    def __init__ (self, width=0, height=0, **options):
        """Class constructor.

        Parameters:
            width -- box' width (default: 0).

            height -- box' height (default: 0).

            options -- keyword arguments.

        Supported keyword arguments:
            -- none at this stage --

        Notice: options are saved into self.options member attribute.
        """
        try:
            (self.width, self.height) = width
        except:
            (self.width, self.height) = (width, height)
        # end try
        self.options = options
    # end def

    def __repr__ (self):
        """Object string representation."""
        return str(self.to_tuple())
    # end def

    @property
    def center_x (self):
        """Center of X dimension - read-only attribute.

        Current implementation merely returns self.width//2.
        """
        return self.width//2
    # end def

    @property
    def center_xy (self):
        """Tuple center of (X, Y) dimensions - read-only attribute.

        Current implementation merely returns a tuple of
        (self.center_x, self.center_y).
        """
        return (self.center_x, self.center_y)
    # end def

    @property
    def center_y (self):
        """Center of Y dimension - read-only attribute.

        Current implementation merely returns self.height//2.
        """
        return self.height//2
    # end def

    @property
    def height (self):
        """Box height - read-write attribute."""
        return self.__height
    # end def

    @height.setter
    def height (self, value):
        self.__height = self.normalize_dimension(value)
    # end def

    def normalize_dimension (self, value):
        """Format value in order to fit with dimension constraints.

        Parameter:
            value -- should be a numeric value (int or float).

        Will raise **ValueError** if `value` is not a numeric value.

        Returns 0 if bool(value) is False.
        """
        if value:
            if isinstance(value, (int, float)):
                return abs(value)
            else:
                raise ValueError(
                    "Value must be of int or float type, not '{}'."
                    .format(value)
                )
            # end if
        # end if
        return 0
    # end def

    @property
    def size (self):
        """Box size - read-only attribute.

        Returns (width, height) tuple.
        """
        return (self.width, self.height)
    # end def

    def to_dict (self):
        """Dictionary representation of object attributes.

        Returns dict(width=self.width, height=self.height).
        """
        return dict(width=self.width, height=self.height)
    # end def

    def to_tuple (self):
        """Tuple representation of object attributes.

        Returns (width, height) tuple.
        """
        return (self.width, self.height)
    # end def

    @property
    def width (self):
        """Box width - read-write attribute."""
        return self.__width
    # end def

    @width.setter
    def width (self, value):
        self.__width = self.normalize_dimension(value)
    # end def

# end class


class WorldBoxBase3D (WorldBoxBase2D):
    """Base class for boxes in a 3D world."""

    def __init__ (self, width=0, height=0, depth=0, **options):
        """Class constructor.

        Parameters:
            width -- box' width (default: 0).

            height -- box' height (default: 0).

            depth -- box' depth (default: 0).

            options -- keyword arguments.

        Supported keyword arguments:
            -- none at this stage --

        Notice: options are saved into self.options member attribute.
        """
        try:
            (self.width, self.height, self.depth) = width
        except:
            self.width = width
            self.height = height
            self.depth = depth
        # end try
        self.options = options
    # end def

    def center (self, query=None):
        """Build tuple of center dimensions along with query.

        Parameter:
            query -- string of dims (default: None).

        If None or omitted, `query` will be replaced by "xyz" string of
        chars.

        Any combination of dims is allowed e.g. "x", "xy", "xz", "yz",
        "xyz", "yxz", "zxy", "zyx" and so on.

        Will always return a tuple object, even for only one dimension
        query e.g. self.center("x") --> (self.center_x, ).
        """
        query = str(query or "xyz").lower()
        _result = []
        if "x" in query:
            _result.append(self.center_x)
        # end if
        if "y" in query:
            _result.append(self.center_y)
        # end if
        if "z" in query:
            _result.append(self.center_z)
        # end if
        return tuple(_result)
    # end def

    @property
    def center_xyz (self):
        """Tuple center of (X, Y, Z) dimensions - read-only attribute.

        Current implementation merely returns a tuple of
        (self.center_x, self.center_y, self.center_z).
        """
        return (self.center_x, self.center_y, self.center_z)
    # end def

    @property
    def center_z (self):
        """Center of Z dimension - read-only attribute.

        Current implementation merely returns self.depth//2.
        """
        return self.depth//2
    # end def

    @property
    def depth (self):
        """Box depth - read-write attribute."""
        return self.__depth
    # end def

    @depth.setter
    def depth (self, value):
        self.__depth = self.normalize_dimension(value)
    # end def

    @property
    def size (self):
        """Box size - read-only attribute.

        Returns (width, height, depth) tuple.
        """
        return (self.width, self.height, self.depth)
    # end def

    def to_dict (self):
        """Dictionary representation of object attributes.

        Returns dict(width=self.width, height=self.height,
        depth=self.depth).
        """
        return dict(
            width=self.width, height=self.height, depth=self.depth
        )
    # end def

    def to_tuple (self):
        """Tuple representation of object attributes.

        Returns (width, height, depth) tuple.
        """
        return (self.width, self.height, self.depth)
    # end def

# end class
