#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Matrix module.

Implements:

* class WorldMatrix -- base class for affine transformation matrices.
* class WorldMatrixError -- matrix generic error handler.
"""

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import operator


class WorldMatrix:
    """Base class for affine transformation matrices.

    This class can be used as is for basic matrix operations (matrix
    sum, matrix multiplication, Hadamard matrix product, scalar matrix
    product, etc).
    """

    def __add__ (self, other):
        """Entrywise sum.

        Returns a matrix object representing a matrix sum.

        May return a submatrix if 'other' matrix size is different than
        'self' matrix size.

        In this case, submatrix will be of minimum size i.e.
        (min(self.rows, other.rows), min(self.columns, other.columns))
        size.
        """
        try:
            return self.__class__(
                tuple(map(operator.add, *rows))
                for rows in zip(self, other)
            )
        except TypeError:
            return NotImplemented
        # end try
    # end def

    def __getitem__ (self, key):
        """Make object iterable."""
        return self.entries[key]
    # end def

    def __init__ (self, sequence=None):
        """Class constructor.

        Parameter:
            sequence -- list of lists of data to be parsed as
            Matrix(rows, columns) entries (default: None). If omitted
            or None, will create an empty matrix.
        """
        self.entries = sequence
    # end def

    def __len__ (self):
        return len(self.entries)
    # end def

    def __mul__ (self, other):
        """Scalar or matrix multiplication.

        Parameter:
            other -- can be a scalar (number) or a matrix to multiply
            with 'self'.

        Returns a new matrix object which represents the rightward
        multiplication of 'self' by 'other' i.e. M = self * other and
        *NOT* M = other * self (non-commutative multiplication).
        """
        if isinstance(other, (int, float)):
            return self.scalar_multiplication(self, other)
        else:
            return self.matrix_multiplication(self, other)
        # end if
    # end def

    def __radd__ (self, other):
        """Reverse entrywise sum.

        Returns an object along with 'other' class type representing a
        matrix sum.

        May return a submatrix if 'other' matrix size is different than
        'self' matrix size.

        In this case, submatrix will be of minimum size i.e.
        (min(self.rows, other.rows), min(self.columns, other.columns))
        size.
        """
        try:
            return other.__class__(
                other.__class__(map(operator.add, *rows))
                for rows in zip(other, self)
            )
        except TypeError:
            return NotImplemented
        # end try
    # end def

    def __repr__ (self):
        """Object string representation."""
        return str(self.entries)
    # end def

    def __rmul__ (self, other):
        """Reverse scalar multiplication.

        Parameter:
            other -- can be a scalar (number) or a matrix to multiply
            with 'self'.

        Returns a new matrix object which represents the leftward
        multiplication of 'self' by 'other' i.e. M = other * self and
        *NOT* M = self * other (non-commutative multiplication).
        """
        if isinstance(other, (int, float)):
            return self.scalar_multiplication(self, other)
        else:
            return self.matrix_multiplication(other, self)
        # end if
    # end def

    def __rsub__ (self, other):
        """Reverse entrywise substraction.

        Returns an object instance along with 'other' class type
        representing a matrix substraction.

        May return a submatrix if 'other' matrix size is different than
        'self' matrix size.

        In this case, submatrix will be of minimum size i.e.
        (min(self.rows, other.rows), min(self.columns, other.columns))
        size.
        """
        try:
            return other.__class__(
                other.__class__(map(operator.sub, *rows))
                for rows in zip(other, self)
            )
        except TypeError:
            return NotImplemented
        # end try
    # end def

    def __sub__ (self, other):
        """Entrywise substraction.

        Returns a matrix object representing a matrix substraction.

        May return a submatrix if 'other' matrix size is different than
        'self' matrix size.

        In this case, submatrix will be of minimum size i.e.
        (min(self.rows, other.rows), min(self.columns, other.columns))
        size.
        """
        try:
            return self.__class__(
                tuple(map(operator.sub, *rows))
                for rows in zip(self, other)
            )
        except TypeError:
            return NotImplemented
        # end try
    # end def

    def cdot (self, matrix):
        """Hadamard matrix product.

        Parameter:
            matrix -- matrix object instance to multiply with 'self'.

        Returns a new matrix object instance representing the Hadamard
        matrix product of 'self' by 'matrix'.
        """
        return self.hadamard_product(self, matrix)
    # end def

    @property
    def columns (self):
        """Matrix' column size - read-only attribute.

        Returns number of columns in matrix (integer).
        """
        return self.entries and len(self.entries[0]) or 0
    # end def

    @property
    def entries (self):
        """Matrix' entries - read-write attribute.

        This attribute is a list of lists of data, as expected for
        Matrix(rows, columns) entries e.g. let an m rows by n columns
        matrix M of e(i,j) entries be::

            M(m, n) = [
                [e(1,1), e(1,2), ..., e(1,n)],
                [e(2,1), e(2,2), ..., e(2,n)],
                ...,
                [e(m,1), e(m,2), ..., e(m,n)]
            ]
        """
        return self.__entries
    # end def

    @entries.setter
    def entries (self, value):
        self.__entries = self.to_tuple(value or [])
    # end def

    @staticmethod
    def hadamard_product (matrix1, *matrices):
        """Hadamard matrix product.

            Parameter:
                matrix1 -- matrix object instance to multiply by other
                matrix object instances.

                matrices -- ordered list of matrix object instances to
                multiply with matrix1 e.g. (matrix2, ..., matrixN).

            Returns a new matrix object instance along with 'matrix1'
            class type, representing the Hadamard matrix product of
            given 'matrix1' by 'matrix2' then 'matrix3' and so on.

            May raise a **WorldMatrixError** if matrices are of
            different size.
        """
        _left_matrix = matrix1
        if _left_matrix:
            _size = lambda m: (len(m), m and len(m[0]) or 0)
            _item_type = list if isinstance(matrix1, list) else tuple
            for _right_matrix in matrices:
                if _size(_left_matrix) == _size(_right_matrix):
                    _left_matrix = tuple(
                        _item_type(map(operator.mul, *_rows))
                        for _rows in zip(_left_matrix, _right_matrix)
                    )
                else:
                    raise WorldMatrixError(
                        "Cannot apply Hadamard product to "
                        "matrices of different size."
                    )
                    return None
                # end if
            # end for
        # end if
        return matrix1.__class__(_left_matrix)
    # end def

    @classmethod
    def I (cls, size, init_value=1):
        """Alias for matrix_identity() class method (uppercase i).

        Parameters:
            cls -- class type to use when instantiating new matrix.

            size -- number of matrix rows and columns (integer).

            init_value -- initial value to fill diagonal entries with
            (default: 1).

        Returns a new identity square matrix object instance of 'size'
        rows and 'size' columns, instantiated along with 'cls' class
        type and filled up with 'init_value' in its diagonal entries.

        For example::

            WorldMatrix.I(3) --> Matrix((1, 0, 0), (0, 1, 0), (0, 0, 1))

            WorldMatrix.I(2, 4) --> Matrix((4, 0), (0, 4))
            is a shortcut for 4 * WorldMatrix.I(2)
        """
        return cls.matrix_identity(size, init_value)
    # end def

    @classmethod
    def matrix_identity (cls, size, init_value=1):
        """New identity matrix instance.

        Parameters:
            cls -- class type to use when instantiating new matrix.

            size -- number of matrix rows and columns (integer).

            init_value -- initial value to fill diagonal entries with
            (default: 1).

        Returns a new identity square matrix object instance of 'size'
        rows and 'size' columns, instantiated along with 'cls' class
        type and filled up with 'init_value' in its diagonal entries.

        For example::

            matrix_identity(3) --> Matrix((1, 0, 0), (0, 1, 0), (0, 0, 1))

            matrix_identity(2, 4) --> Matrix((4, 0), (0, 4))
            is a shortcut for 4 * WorldMatrix.matrix_identity(2)
        """
        size = abs(int(size))
        _sequence = tuple(
            tuple(int(i == j) and init_value for j in range(size))
            for i in range(size)
        )
        return cls(_sequence)
    # end def

    @staticmethod
    def matrix_multiplication (matrix1, *matrices):
        """Matrix multiplication.

            Parameter:
                matrix1 -- matrix object instance to multiply by other
                matrix object instances.

                matrices -- ordered list of matrix object instances to
                multiply with matrix1 e.g. (matrix2, ..., matrixN).

            Returns a new matrix object instance along with 'matrix1'
            class type, representing the matrix multiplication of given
            'matrix1' by 'matrix2' then 'matrix3' and so on, in this
            particular order i.e. M = matrix1 * matrix2 * ... * matrixN
            and *NOT* M = matrixN * ... * matrix2 * matrix1
            (non-commutative matrix multiplication).

            May raise **WorldMatrixError** if a left matrix number of
            columns is different than right matrix number of rows.
        """
        _left_matrix = matrix1
        if _left_matrix:
            _item_type = list if isinstance(matrix1, list) else tuple
            for _right_matrix in matrices:
                if len(_left_matrix[0]) == len(_right_matrix):
                    _left_matrix = tuple(
                        _item_type(
                            sum(map(operator.mul, _lrow, _rcol))
                            for _rcol in zip(*_right_matrix)
                        )
                        for _lrow in _left_matrix
                    )
                else:
                    raise WorldMatrixError(
                        "Cannot multiply matrices: "
                        "left matrix number of columns is "
                        "different than right matrix number of rows."
                    )
                    return None
                # end if
            # end for
        # end if
        return matrix1.__class__(_left_matrix)
    # end def

    @classmethod
    def matrix_new (cls, rows, columns, init_value=0):
        """New matrix instance.

        Parameters:
            cls -- class type to use when instantiating new matrix.

            rows -- number of matrix rows (integer).

            columns -- number of matrix columns (integer).

            init_value -- initial value to fill all matrix entries with
            (default: 0).

        Returns a new matrix object instance along with 'cls' class
        type filled up with 'rows' by 'columns' entries of 'init_value'
        value.
        """
        rows = abs(int(rows))
        columns = abs(int(columns))
        _sequence = ((init_value,) * columns for n in range(rows))
        return cls(_sequence)
    # end def

    @staticmethod
    def matrix_to_type (iterable_type, sequence):
        """Convert a given sequence to an iterable type list of
        iterable types.

        Parameters:
            iterable_type -- iterable type to apply to given sequence.

            sequence -- list of lists of data to be parsed as
            Matrix(rows, columns) entries.

        May raise a **TypeError** if 'sequence' is not some kind of
        iterable of iterables, as expected.

        May raise **WorldMatrixError** if 'sequence' contains lists of
        several lengths i.e. inconsistent matrix' column width.
        """
        _rows = []
        _columns = 0
        for _row in sequence:
            _columns = _columns or len(_row)
            if _columns == len(_row):
                _rows.append(iterable_type(_row))
            else:
                raise WorldMatrixError(
                    "Sequence has several column widths, "
                    "which is *NOT* allowed into a matrix structure."
                )
                return None
            # end if
        # end for
        return iterable_type(_rows)
    # end def

    @staticmethod
    def matrix_transpose (matrix):
        """Matrix transpose.

        Parameter:
            matrix -- matrix object to transpose.

        Returns a new matrix object instance along with 'matrix' class
        type, representing the matrix transpose of given 'matrix' i.e.
        rows become columns and columns become rows.
        """
        return matrix.__class__(zip(*matrix))
    # end def

    @property
    def rows (self):
        """Matrix' row size - read-only attribute.

        Returns number of rows in matrix (integer).
        """
        return len(self.entries)
    # end def

    @staticmethod
    def scalar_multiplication (matrix, scalar):
        """Scalar multiplication of a matrix by a number.

            Parameters:
                matrix -- matrix object instance to multiply by scalar
                number.

                scalar -- scalar number (int or float) used to multiply
                matrix' entries.

            Returns a new matrix object instance along with 'matrix'
            class type, representing the scalar multiplication of given
            'matrix' by 'scalar' number.
        """
        return matrix.__class__(
            tuple(entry * scalar for entry in row) for row in matrix
        )
    # end def

    @property
    def size (self):
        """Matrix' size - read-only attribute.

        Returns a (self.rows, self.columns) tuple.
        """
        return (self.rows, self.columns)
    # end def

    def to_list (self, sequence=None):
        """Convert a given sequence to a list of lists.

        Parameter:
            sequence -- list of lists of data to be parsed as
            Matrix(rows, columns) entries (default: None). If omitted
            or None, will use self.entries instead.

        Returns a Python list() object of list() objects.

        May raise a **TypeError** if 'sequence' is not some kind of
        iterable of iterables, as expected.

        May raise **WorldMatrixError** if 'sequence' contains lists of
        several lengths i.e. inconsistent matrix' column width.
        """
        if sequence is None:
            sequence = self.entries
        # end if
        return self.matrix_to_type(list, sequence)
    # end def

    def to_tuple (self, sequence=None):
        """Convert a given sequence to a tuple of tuples.

        Parameter:
            sequence -- list of lists of data to be parsed as
            Matrix(rows, columns) entries (default: None). If omitted
            or None, will use self.entries instead.

        Returns a Python tuple() object of tuple() objects.

        May raise a **TypeError** if 'sequence' is not some kind of
        iterable of iterables, as expected.

        May raise **WorldMatrixError** if 'sequence' contains lists of
        several lengths i.e. inconsistent matrix' column width.
        """
        if sequence is None:
            sequence = self.entries
        # end if
        return self.matrix_to_type(tuple, sequence)
    # end def

    def transpose (self, matrix=None):
        """Matrix transpose.

        Parameter:
            matrix -- matrix object to transpose (default: None). If
            omitted or None, will be replaced by 'self'.

        Returns a new matrix object instance along with 'matrix' class
        type, representing the matrix transpose of given 'matrix' i.e.
        rows become columns and columns become rows.
        """
        return self.matrix_transpose(matrix or self)
    # end def

# end class


class WorldMatrixError (Exception):
    """Matrix general error handler."""
    pass
# end class
