#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Geometrical point module.

Implements:

* class WorldPoint2D -- geometrical point for 2D world.
* class WorldPoint3D -- geometrical point for 3D world.
"""

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import operator


class WorldPoint2D:
    """Geometrical point for 2D world."""

    def __add__ (self, other):
        """Analytical sum of two points (coordinates sum)."""
        try:
            return self.__class__(*map(operator.add, self, other))
        except TypeError:
            return NotImplemented
        # end try
    # end def

    def __floordiv__ (self, other):
        """Floor division (//) of point's coordinates by a number."""
        if isinstance(other, (int, float)):
            return self.__class__(*(dim // other for dim in self))
        else:
            return NotImplemented
        # end if
    # end def

    def __getitem__ (self, key):
        """Make object iterable."""
        return self.to_tuple()[key]
    # end def

    def __init__ (self, x=0, y=0):
        """Class constructor.

        Parameters:
            x -- numeric dimension coordinate (default: 0).

            y -- numeric dimension coordinate (default: 0).
        """
        try:
            (self.x, self.y) = x
        except:
            (self.x, self.y) = (x, y)
        # end try
    # end def

    def __mul__ (self, other):
        """Multiplication of point's coordinates by a number."""
        if isinstance(other, (int, float)):
            return self.__class__(*(dim * other for dim in self))
        else:
            return NotImplemented
        # end if
    # end def

    def __repr__ (self):
        """Object string representation."""
        return str(self.to_tuple())
    # end def

    def __rmul__ (self, other):
        """Reverse multiplication - see __mul__() for more detail."""
        return self.__mul__(other)
    # end def

    def __sub__ (self, other):
        """Analytical sub of two points (coordinates substraction)."""
        try:
            return self.__class__(*map(operator.sub, self, other))
        except TypeError:
            return NotImplemented
        # end try
    # end def

    def __truediv__ (self, other):
        """True division (/) of point's coordinates by a number."""
        if isinstance(other, (int, float)):
            return self.__class__(*(dim / other for dim in self))
        else:
            return NotImplemented
        # end if
    # end def

    def cdot (self, other):
        """Hadamard matrix product.

        Examples::

            Point(x, y).cdot(Point(a, b)) == Point(x*a, y*b)

            Point(x, y, z).cdot(Point(a, b, c)) == Point(x*a, y*b, z*c)
        """
        return self.__class__(*map(operator.mul, self, other))
    # end def

    def normalize_dimension (self, value):
        """Format value in order to fit with dimension constraints.

        Parameter:
            value -- should be a numeric value (int or float).

        Will raise **ValueError** if `value` is not a numeric value.

        Returns 0 if bool(value) is False.
        """
        if value:
            if isinstance(value, (int, float)):
                return value
            else:
                raise ValueError(
                    "Value must be of int or float type, not '{}'."
                    .format(value)
                )
            # end if
        # end if
        return 0
    # end def

    def to_dict (self):
        """Dictionary representation of object attributes.

        Returns dict(x=self.x, y=self.y).
        """
        return dict(x=self.x, y=self.y)
    # end def

    def to_tuple (self):
        """Tuple representation of object attributes.

        Returns (x, y) tuple.
        """
        return self.xy
    # end def

    @property
    def x (self):
        """x dimension coordinate - read-write attribute."""
        return self.__x
    # end def

    @x.setter
    def x (self, value):
        self.__x = self.normalize_dimension(value)
    # end def

    @property
    def xy (self):
        """(x, y) coordinates - read-only attribute."""
        return (self.x, self.y)
    # end def

    @property
    def y (self):
        """y dimension coordinate - read-write attribute."""
        return self.__y
    # end def

    @y.setter
    def y (self, value):
        self.__y = self.normalize_dimension(value)
    # end def

# end class


class WorldPoint3D (WorldPoint2D):
    """Geometrical point for 3D world."""

    def __init__ (self, x=0, y=0, z=0):
        """Class constructor.

        Parameters:
            x -- numeric dimension coordinate (default: 0).

            y -- numeric dimension coordinate (default: 0).

            z -- numeric dimension coordinate (default: 0).
        """
        try:
            (self.x, self.y, self.z) = x
        except:
            (self.x, self.y, self.z) = (x, y, z)
        # end try
    # end def

    def to_dict (self):
        """Dictionary representation of object attributes.

        Returns dict(x=self.x, y=self.y, z=self.z).
        """
        return dict(x=self.x, y=self.y, z=self.z)
    # end def

    def to_tuple (self):
        """Tuple representation of object attributes.

        Returns (x, y, z) tuple.
        """
        return self.xyz
    # end def

    @property
    def xyz (self):
        """(x, y, z) coordinates - read-only attribute."""
        return (self.x, self.y, self.z)
    # end def

    @property
    def z (self):
        """z dimension coordinate - read-write attribute."""
        return self.__z
    # end def

    @z.setter
    def z (self, value):
        self.__z = self.normalize_dimension(value)
    # end def

# end class
