#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Geometrical vector module.

Implements:

* class WorldVector2D -- geometrical vector for 2D world.
* class WorldVector3D -- geometrical vector for 3D world.
"""

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

try:
    from .point import *
except:
    from point import *
# end try


class WorldVector2D (WorldPoint2D):
    """Geometrical vector for 2D world."""

    def __abs__ (self):
        """Vector norm."""
        return sum(dim**2 for dim in self)**0.5
    # end def

    def _dim_query (self, query, prefix):
        """Protected method def - for internal use only."""
        for _dim in query:
            exec(
                "self.{prefix}_{dimension}()"
                .format(prefix=prefix, dimension=_dim)
            )
        # end for
        return self
    # end def

    def abs (self, query=None):
        """Force vector coordinates to abs() along with query.

        Parameter:
            query -- string of dims (default: None).

        If None or omitted, `query` will be replaced by "xy" string of
        chars.

        Any combination of dims is allowed e.g. "x", "xy", "yx", "y".

        Returns self.
        """
        query = str(query or "xy").lower()
        return self._dim_query(query, "abs")
    # end def

    def abs_x (self):
        """Force vector x coordinate to abs(x)."""
        self.x = abs(self.x)
        return self
    # end def

    def abs_y (self):
        """Force vector y coordinate to abs(y)."""
        self.y = abs(self.y)
        return self
    # end def

    def flip (self, query=None):
        """Flip (invert) vector coordinates along with query.

        Parameter:
            query -- string of dims (default: None).

        If None or omitted, `query` will be replaced by "xy" string of
        chars.

        Any combination of dims is allowed e.g. "x", "xy", "yx", "y".

        Returns self.
        """
        query = str(query or "xy").lower()
        return self._dim_query(query, "flip")
    # end def

    def flip_x (self):
        """Flip vector x coordinate to -x."""
        self.x = -self.x
        return self
    # end def

    def flip_y (self):
        """Flip vector y coordinate to -y."""
        self.y = -self.y
        return self
    # end def

    def neg (self, query=None):
        """Negate vector coordinates along with query.

        Parameter:
            query -- string of dims (default: None).

        If None or omitted, `query` will be replaced by "xy" string of
        chars.

        Any combination of dims is allowed e.g. "x", "xy", "yx", "y".

        Returns self.
        """
        query = str(query or "xy").lower()
        return self._dim_query(query, "neg")
    # end def

    def neg_x (self):
        """Force vector x coordinate to -abs(x)."""
        self.x = -abs(self.x)
        return self
    # end def

    def neg_y (self):
        """Force vector y coordinate to -abs(y)."""
        self.y = -abs(self.y)
        return self
    # end def

# end class


class WorldVector3D (WorldPoint3D, WorldVector2D):
    """Geometrical vector for 3D world."""

    def __init__ (self, x=0, y=0, z=0):
        """Class constructor.

        Parameters:
            x -- numeric dimension coordinate (default: 0).

            y -- numeric dimension coordinate (default: 0).

            z -- numeric dimension coordinate (default: 0).
        """
        WorldPoint3D.__init__(self, x, y, z)
    # end def

    def abs (self, query=None):
        """Force vector coordinates to abs() along with query.

        Parameter:
            query -- string of dims (default: None).

        If None or omitted, `query` will be replaced by "xyz" string of
        chars.

        Any combination of dims is allowed e.g. "x", "xy", "xz", "yz",
        "xyz", "yxz", "zxy", "zyx" and so on.

        Returns self.
        """
        return super().abs(query or "xyz")
    # end def

    def abs_z (self):
        """Force vector z coordinate to abs(z)."""
        self.z = abs(self.z)
        return self
    # end def

    def flip (self, query=None):
        """Flip (invert) vector coordinates along with query.

        Parameter:
            query -- string of dims (default: None).

        If None or omitted, `query` will be replaced by "xyz" string of
        chars.

        Any combination of dims is allowed e.g. "x", "xy", "xz", "yz",
        "xyz", "yxz", "zxy", "zyx" and so on.

        Returns self.
        """
        return super().flip(query or "xyz")
    # end def

    def flip_z (self):
        """Flip vector z coordinate to -z."""
        self.z = -self.z
        return self
    # end def

    def neg (self, query=None):
        """Negate vector coordinates along with query.

        Parameter:
            query -- string of dims (default: None).

        If None or omitted, `query` will be replaced by "xyz" string of
        chars.

        Any combination of dims is allowed e.g. "x", "xy", "xz", "yz",
        "xyz", "yxz", "zxy", "zyx" and so on.

        Returns self.
        """
        return super().neg(query or "xyz")
    # end def

    def neg_z (self):
        """Force vector z coordinate to -abs(z)."""
        self.z = -abs(self.z)
        return self
    # end def

# end class
